/* Highly inspirated by nehe lesson 4 port to GX by WinterMute. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <gccore.h>
#include <wiiuse/wpad.h>

#include "geom.h"
#include "cube_data.h"


#define DEFAULT_FIFO_SIZE (256 * 1024)

/* Pointers to the allocated framebuffer memory areas. */
static void *frameBuffer[2] = { NULL, NULL} ;
GXRModeObj *rmode ;




void render_shape ()
{
  int i ;
  struct vector *tmp ;

  /* Last argument is the vertices count. */
  GX_Begin (GX_TRIANGLES, GX_VTXFMT0, (NB_TRIANGLES * 3)) ;

  for (i = 0; i < NB_TRIANGLES; i++) {
    tmp = &vertices[triangles[i].p0] ;
    GX_Position3f32 (tmp->x, tmp->y, tmp->z) ;
    GX_Color3f32 (1.0f, 0.0f, 0.0f) ;         /* Red. */

    tmp = &vertices[triangles[i].p1] ;
    GX_Position3f32 (tmp->x, tmp->y, tmp->z) ;
    GX_Color3f32 (0.0f, 1.0f, 0.0f) ;         /* Green. */

    tmp = &vertices[triangles[i].p2] ;
    GX_Position3f32 (tmp->x, tmp->y, tmp->z) ;
    GX_Color3f32 (0.0f, 0.0f, 1.0f) ;         /* Blue. */
  }
  GX_End () ;
}




int main (int argc, char **argv)
{
  f32 yscale ;
  u32 xfbHeight ;
  Mtx view,perspective ; /* View and perspective matrices. */
  Mtx model, modelview ;
  float cube_rotation_around_x_angle = 0.0f ;
  float cube_rotation_around_z_angle = 0.0f ;
  u32 fb = 0 ;          /* Initial framebuffer index. */
  GXColor background = {0, 0, 0, 0xff} ;
  void *gp_fifo = NULL;

  /* Initialize the Wii. */
  VIDEO_Init () ;
  WPAD_Init () ;

  /* Say we want to get buttons, acceleromater and IR-tracking info. */
  WPAD_SetDataFormat (WPAD_CHAN_0, WPAD_FMT_BTNS_ACC_IR) ;

  rmode = VIDEO_GetPreferredMode (NULL) ;

  /* Allocate 2 framebuffers for double buffering. */
  frameBuffer[0] = MEM_K0_TO_K1 (SYS_AllocateFramebuffer (rmode)) ;
  frameBuffer[1] = MEM_K0_TO_K1 (SYS_AllocateFramebuffer (rmode)) ;

  VIDEO_Configure(rmode) ;
  /* Switch to initial framebuffer. */
  VIDEO_SetNextFramebuffer (frameBuffer[fb]) ;
  VIDEO_SetBlack (FALSE) ;                   /* Make the display visible. */
  VIDEO_Flush () ;
  VIDEO_WaitVSync () ;                /* Wait for the next vertical retrace. */
  if (rmode->viTVMode & VI_NON_INTERLACE) VIDEO_WaitVSync () ;

  /* Setup the fifo and then init the flipper. */
  gp_fifo = memalign (32, DEFAULT_FIFO_SIZE) ;
  memset (gp_fifo, 0, DEFAULT_FIFO_SIZE) ;
  GX_Init (gp_fifo, DEFAULT_FIFO_SIZE) ;

  /* Clears the bg to color and clears the z buffer. */
  GX_SetCopyClear (background, 0x00ffffff) ;

  /* Other gx setup. */
  /*              xOrig, yOrig, width, height, nearZ,farZ */
  GX_SetViewport (0, 0, rmode->fbWidth, rmode->efbHeight, 0, 1) ;
  /* Set scale vertical ratio between "embedded" (i.e. source) */
  /* and "external" (destination) framebuffers.                */
  yscale = GX_GetYScaleFactor (rmode->efbHeight, rmode->xfbHeight) ;
  xfbHeight = GX_SetDispCopyYScale (yscale) ;
  GX_SetScissor (0, 0, rmode->fbWidth, rmode->efbHeight) ;  /* Clipping ? */
  /* Describes the source image area in the EFB. */
  GX_SetDispCopySrc (0, 0, rmode->fbWidth, rmode->efbHeight) ;
  /* Describes how to map the source image into the XFB. */
  GX_SetDispCopyDst (rmode->fbWidth, xfbHeight) ;
  /* ??? */
  GX_SetCopyFilter (rmode->aa, rmode->sample_pattern, GX_TRUE, rmode->vfilter) ;
  /* ??? */
  GX_SetFieldMode
    (rmode->field_rendering,
     ((rmode->viHeight == 2 * rmode->xfbHeight) ? GX_ENABLE:GX_DISABLE)) ;

  GX_SetCullMode (GX_CULL_NONE) ;
  /* Copy the EFB to XFB. Flag tells if the color and Z buffers will */
  /* be cleared during the copy using values of GX_SetCopyClear).    */
  GX_CopyDisp (frameBuffer[fb], GX_TRUE) ;
  /* Sets gamma correction applied to pixels during EFB to XFB copy. */
  GX_SetDispCopyGamma (GX_GM_1_0) ;


  /* Setup the vertex descriptor to tell the flipper to expect direct data. */
  GX_ClearVtxDesc () ;
  GX_SetVtxDesc (GX_VA_POS, GX_DIRECT) ;
  GX_SetVtxDesc (GX_VA_CLR0, GX_DIRECT) ;

  /* setup the vertex attribute table
     describes the data
     args: vat location 0-7, type of data, data format, size, scale
     so for ex. in the first call we are sending position data with
     3 values X,Y,Z of size F32. scale sets the number of fractional
     bits for non float data. */
  GX_SetVtxAttrFmt (GX_VTXFMT0, GX_VA_POS, GX_POS_XYZ, GX_F32, 0);
  GX_SetVtxAttrFmt (GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGB8, 0);

  GX_SetNumChans(1) ;   /* ??? */
  GX_SetNumTexGens (0) ;   /* No texture? ??? */
  GX_SetTevOrder (GX_TEVSTAGE0, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR0A0) ;
  GX_SetTevOp (GX_TEVSTAGE0, GX_PASSCLR) ;


  /* Setup our camera at the origin looking down the -z axis with y up. */
  Vector cam = {0.0F, 0.0F, 0.0F} ;
  Vector up = {0.0F, 1.0F, 0.0F} ;
  Vector look = {0.0F, 0.0F, -1.0F} ;
  guLookAt (view, &cam, &up, &look) ;

  /* Setup our projection matrix. This creates a perspective matrix with a
     view angle of 90, and aspect ratio based on the display resolution. */
  f32 w = rmode->viWidth ;
  f32 h = rmode->viHeight ;
  guPerspective (perspective, 45, (f32) w / h, 0.1F, 300.0F) ;
  GX_LoadProjectionMtx (perspective, GX_PERSPECTIVE) ;

  Vector Xaxis = {1,0,0} ;
  Vector Zaxis = {0,0,1} ;
  struct orient_t wiimote0_orient ;
  Mtx tmp_matrix ;

  while(1) {
    WPAD_ScanPads () ;
    if (WPAD_ButtonsDown (WPAD_CHAN_0) & WPAD_BUTTON_HOME) exit(0) ;

    /* Do this before drawing. */
    GX_SetViewport (0, 0, rmode->fbWidth, rmode->efbHeight, 0, 1) ;

    guMtxIdentity (model) ;
    // guMtxRotAxisDeg (model, &Xaxis, cube_rotation_around_x_angle) ;
    // guMtxRotAxisDeg (model, &Zaxis, cube_rotation_around_z_angle) ;
    guMtxRotAxisDeg (tmp_matrix, &Xaxis, cube_rotation_around_x_angle) ;
    guMtxConcat (model, tmp_matrix, model) ;  /* model0, tmp -> model0. */
    guMtxRotAxisDeg (tmp_matrix, &Zaxis, cube_rotation_around_z_angle) ;
    guMtxConcat (model, tmp_matrix, model) ;  /* model0, tmp -> model0. */

    /* Translate the model in front of us (at -6 along z). */
    guMtxTransApply (model, model, 0.0f, 0.0f, -6.0f) ;

    guMtxConcat (view, model, modelview) ;
    /* Load the modelview matrix into matrix memory. */
    GX_LoadPosMtxImm (modelview, GX_PNMTX0) ;

    /* Render the shape. */
    render_shape () ;

    /* Do this stuff after drawing. */
    GX_DrawDone () ;

    fb ^= 1 ;                /* Flip framebuffer. */
    GX_SetZMode (GX_TRUE, GX_LEQUAL, GX_TRUE) ;         /* ??? */
    GX_SetColorUpdate (GX_TRUE) ;                       /* ??? */
    /* Copy the EFB to XFB. Flag tells if the color and Z buffers will
       be cleared during the copy using values of GX_SetCopyClear. */
    GX_CopyDisp (frameBuffer[fb],GX_TRUE) ;

    VIDEO_SetNextFramebuffer (frameBuffer[fb]) ;

    VIDEO_Flush () ;
    VIDEO_WaitVSync () ;

    /* Ask info about the orientation of the wiimote. */
    WPAD_Orientation (WPAD_CHAN_0, &wiimote0_orient) ;
    /* Rotate amount around x is the up/down slope. */
    cube_rotation_around_x_angle -= (wiimote0_orient.pitch / 10.0) ;
    /* Rotate amount around z is the left/right slope. */
    cube_rotation_around_z_angle -= (wiimote0_orient.roll / 10.0) ;
  }
  return (0) ;
}
